import pytest
import datetime
import kenpompy.team as kpteam
import kenpompy.misc as kpmisc

def test_get_valid_teams(browser):
	expected = 357

	teams_2021 = kpteam.get_valid_teams(browser, season = '2021')
	assert len(teams_2021) == expected

	valid_2021_teams = ['Gonzaga', 'Penn St.', 'Florida', 'Xavier', 'VMI', 'Kennesaw St.', 'Wagner', 'Bucknell', 'Maryland Eastern Shore', 'Cal St. Fullerton']
	for team in valid_2021_teams:
		assert team in teams_2021

	invalid_2021_teams = ['Gonpraga', 'North Carolina    A&T', 'Bayton', 'LMU', 'Goopton']
	for team in invalid_2021_teams:
		assert team not in teams_2021

	teams_2003 = kpteam.get_valid_teams(browser, season = '2003')
	assert len(teams_2003) == 327

	valid_2003_teams = ['Kentucky', 'Kansas', 'Georgetown', 'Dayton', 'South Carolina', 'Fresno St.', 'Iowa', 'SMU', 'TCU', 'North Carolina A&T']
	for team in valid_2003_teams:
		assert team in teams_2003
	
	invalid_2003_teams = ['Loyola Marymnt University', 'YRU', 'Praget', 'Invalid U', 'SRTU', 'Kennesaw St.']
	for team in invalid_2003_teams:
		assert team not in teams_2003
	
	teams_2020 = kpteam.get_valid_teams(browser, season = '2020')
	for team in teams_2020:
		assert '*' not in team


def test_get_schedule(browser):
	expected = ['Sat Dec 15', '122', '286', 'Portland St.', 'W, 85-58', '70', 'Away', '10-1', '', 'None']

	df = kpteam.get_schedule(browser, team="Loyola Marymount", season = '2019')
	assert [str(i) for i in df[df.Date == 'Sat Dec 15'].iloc[0].to_list()] == expected
	assert df.shape == (34, 10)

	expected = ['Mon Apr 2', '1', '7', 'Michigan', 'W, 79-62', '67', 'Neutral', '36-4', '', 'NCAA']
	df = kpteam.get_schedule(browser, team='Villanova', season=2018)
	assert df[df.Date == 'Mon Apr 2'].iloc[0].to_list() == expected

	date = datetime.date.today()
	currentYear = kpmisc.get_current_season(browser)
	nextYear = str(int(currentYear)+1)

	with pytest.raises(ValueError):
		kpteam.get_schedule(browser, team="Iowa", season = '1998')

	with pytest.raises(ValueError):
		kpteam.get_schedule(browser, team="Kansas", season = nextYear)

	with pytest.raises(ValueError):
		kpteam.get_schedule(browser, season = "2009")

	with pytest.raises(ValueError):
		kpteam.get_schedule(browser, team='Merrimack', season=2019)

	with pytest.raises(ValueError):
		kpteam.get_schedule(browser, team='Incorrect Team Name', season=2017)

	centenary_expected = ['Sat Nov 11', '', '172', 'TCU', 'L, 72-66', '76', 'Away', '0-1', '', 'None']
	centenary_df = kpteam.get_schedule(browser, team='Centenary', season=2007)
	assert [str(i) for i in centenary_df[centenary_df.Date == 'Sat Nov 11'].iloc[0].to_list()] == centenary_expected
	assert centenary_df.shape == (31, 10)

	with pytest.raises(ValueError):
		kpteam.get_schedule(browser, team='Centenary', season=2017)

	# Make sure that the valid team check is triggered
	with pytest.raises(ValueError):
		kpteam.get_schedule(browser, season = '2013', team="LMU")

def test_get_scouting_report(browser):
	data = kpteam.get_scouting_report(browser, 'Purdue', season=2023)
	assert data['OE'] == 117.7
	assert data['APLD'] == 18.1
	# Not including SoS and Personnel (88 with)
	assert len(data) == 70

	data = kpteam.get_scouting_report(browser, 'Purdue', season=2023, conference_only=True)
	assert data['OE'] == 111.3
	assert data['APLD'] == 18.3
	# Not including SoS and Personnel (88 with)
	assert len(data) == 70

	data = kpteam.get_scouting_report(browser, 'Kansas', season=2009)
	assert data['OE'] == 114.5
	assert data['APLD'] == ''
	# Not including SoS and Personnel (88 with)
	assert len(data) == 70

	data = kpteam.get_scouting_report(browser, 'Kansas', season=2009, conference_only=True)
	assert data['OE'] == 109.5
	assert data['APLD'] == ''
	# Not including SoS and Personnel (88 with)
	assert len(data) == 70



"""
This module contains functions for scraping the team page kenpom.com tables into
pandas dataframes
"""

import pandas as pd
from io import StringIO
from .misc import get_current_season
import re
from cloudscraper import CloudScraper
from bs4 import BeautifulSoup
from codecs import encode, decode
from typing import Optional
from .utils import get_html

def get_valid_teams(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the teams (https://kenpom.com) into a list.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.

	Returns:
		team_list (list): List containing all valid teams for the given season on kenpom.com.
	"""

	url = "https://kenpom.com"
	url = url + '?y=' + str(season)

	teams = BeautifulSoup(get_html(browser, url), "html.parser")
	table = teams.find_all('table')[0]
	team_df = pd.read_html(StringIO(str(table)))
	# Get only the team column.
	team_df = team_df[0].iloc[:, 1]
 	# Remove NCAA tourny seeds for previous seasons.
	team_df = team_df.str.replace(r'\d+\**', '', regex=True)
	team_df = team_df.str.rstrip()
	team_df = team_df.dropna()
	# Remove leftover team headers
	team_list = team_df.values.tolist()
	team_list = [team for team in team_df if team != "Team"]

	return team_list

def get_schedule(browser: CloudScraper, team: Optional[str]=None, season: Optional[str]=None):
	"""
	Scrapes a team's schedule from (https://kenpom.com/team.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function
		team (str, optional): Used to determine which team to scrape for schedule.
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.

	Returns:
		team_df (pandas dataframe): Dataframe containing a team's schedule for the given season.

	Raises:
		ValueError if `season` is less than 1999.
		ValueError if `season` is greater than the current year.
		ValueError if `team` is not in the valid team list.
	"""

	url = 'https://kenpom.com/team.php'
	current_season = get_current_season(browser)

	if season:
		if int(season) < 1999:
			raise ValueError(
				'season cannot be less than 1999, as data only goes back that far.')
		if int(season) > int(current_season):
			raise ValueError(
				'season cannot be greater than the current year.')
	else:
		season = current_season

	if team==None or team not in get_valid_teams(browser, season):
			raise ValueError(
				'the team does not exist in kenpom in the given year.  Check that the spelling matches (https://kenpom.com) exactly.')
	
	# Sanitize team name
	team = team.replace(" ", "+")
	team = team.replace("&", "%26")
	url = url + "?team=" + str(team)
	url = url + "&y=" + str(season)

	schedule = BeautifulSoup(get_html(browser, url), "html.parser")
	table = schedule.find_all('table')[1]
	schedule_df = pd.read_html(StringIO(str(table)))

	# Dataframe Tidying
	schedule_df = schedule_df[0]
	# Teams 2010 and earlier do not show their team rank, add column for consistency
	if(len(schedule_df.columns) == 10):
		schedule_df.insert(1, 'Team Rank', '')
	schedule_df.columns = ['Date', 'Team Rank', 'Opponent Rank', 'Opponent Name', 'Result', 'Possession Number',
					  'A', 'Location', 'Record', 'Conference', 'B']
	schedule_df = schedule_df.drop(columns = ['A', 'B'])
	schedule_df = schedule_df.fillna('')

	# Add postseason tournament info to a distinct column
	schedule_df['Postseason'] = None
	# Enumerate tournament names and their row indices
	postseason_labels = schedule_df[(schedule_df['Team Rank'].str.contains('Tournament')) | (schedule_df['Team Rank'].str.contains('Postseason'))].reset_index()[['index', 'Date']].values.tolist()
	# Tournament name preprocessing
	postseason_labels = list(map(lambda x: [x[0], re.sub(r'(?:\sConference)?\sTournament.*?$', '', x[1])], postseason_labels))
	# Loop tournaments in schedule and apply to associated games
	i = 0
	while i < len(postseason_labels):
		if i != len(postseason_labels) - 1:
			schedule_df.loc[postseason_labels[i][0]:postseason_labels[i+1][0]-1, 'Postseason'] = postseason_labels[i][1]
		else:
			schedule_df.loc[postseason_labels[i][0]:, 'Postseason'] = postseason_labels[i][1]
		i += 1
	# Remove table data not corresponding to a scheduled competition
	schedule_df = schedule_df[schedule_df['Date'] != schedule_df['Result']]
	schedule_df = schedule_df[schedule_df['Date'] != 'Date']

	return schedule_df.reset_index(drop=True)

def get_scouting_report(browser: CloudScraper, team: str, season: Optional[int]=None, conference_only: bool=False):
	"""
    Retrieves and parses team scouting report data from (https://kenpom.com/team.php) into a dictionary.

    Args:
    	browser (CloudScraper): The mechanize browser object for web scraping.
    	team (str): team: Used to determine which team to scrape for schedule.
    	season (int, optional): Used to define different seasons. 1999 is the earliest available season.
    	conference_only (bool, optional): When True, only conference-related stats are retrieved; otherwise, all stats are fetched.

    Returns:
    	dict: A dictionary containing various team statistics.

    Raises:
    	ValueError if the provided season is earlier than 1999 or greater than the current year
		ValueError if the team name is invalid or not found in the specified year
	"""

	url = 'https://kenpom.com/team.php'

	current_season = get_current_season(browser)

	if season:
		if int(season) < 1999:
			raise ValueError(
				'season cannot be less than 1999, as data only goes back that far.')
		if int(season) > current_season:
			raise ValueError(
				'season cannot be greater than the current year.')
	else:
		season = int(current_season)

	if team==None or team not in get_valid_teams(browser, season):
			raise ValueError(
				'the team does not exist in kenpom in the given year.  Check that the spelling matches (https://kenpom.com) exactly.')
	
	# Sanitize team name
	team = team.replace(" ", "+")
	team = team.replace("&", "%26")
	url = url + "?team=" + str(team)
	url = url + "&y=" + str(season)

	report = BeautifulSoup(get_html(browser, url), "html.parser")
	# Find all script tags and filter for ones without src attribute (inline scripts)
	all_scripts = report.find_all("script", { "type": "text/javascript"})
	scouting_report_scripts = None
	for script in all_scripts:
		if not script.get('src'):  # Find script tag without src attribute
			scouting_report_scripts = script
			break

	extraction_pattern = re.compile(r"\$\(\"td#(?P<token>[A-Za-z0-9]+)\"\)\.html\(\"(.+)\"\);")
	if conference_only:
		pattern = re.compile(r"\$\(':checkbox'\).click\(function\(\) \{([^\}]+)}")
	else:
		pattern = re.compile(r"function tableStart\(\) \{([^\}]+)}")

	stats = extraction_pattern.findall(decode(encode(pattern.search(str(scouting_report_scripts.contents[0])).groups()[0], 'latin-1', 'backslashreplace'), 'unicode-escape'))
	stats = list(map(lambda x: (x[0], float(BeautifulSoup(x[1], "lxml").find('a').contents[0]), int(str(BeautifulSoup(x[1], "lxml").find('span', { "class": "seed" }).contents[0]))), stats[2:]))
	# Defaulting each stat to '' for earlier years which might not have all the stats
	stats_df = {'OE': '', 'OE.Rank': '', 'DE': '', 'DE.Rank': '', 'Tempo': '', 'Tempo.Rank': '', 'APLO': '', 'APLO.Rank': '', 'APLD': '', 'APLD.Rank': '', 'eFG': '', 'eFG.Rank': '', 'DeFG': '', 'DeFG.Rank': '', 'TOPct': '', 'TOPct.Rank': '', 'DTOPct': '', 'DTOPct.Rank': '', 'ORPct': '', 'ORPct.Rank': '', 'DORPct': '', 'DORPct.Rank': '', 'FTR': '', 'FTR.Rank': '', 'DFTR': '', 'DFTR.Rank': '', '3Pct': '', '3Pct.Rank': '', 'D3Pct': '', 'D3Pct.Rank': '', '2Pct': '', '2Pct.Rank': '', 'D2Pct': '', 'D2Pct.Rank': '', 'FTPct': '', 'FTPct.Rank': '', 'DFTPct': '', 'DFTPct.Rank': '', 'BlockPct': '', 'BlockPct.Rank': '', 'DBlockPct': '', 'DBlockPct.Rank': '', 'StlRate': '', 'StlRate.Rank': '', 'DStlRate': '', 'DStlRate.Rank': '', 'NSTRate': '', 'NSTRate.Rank': '', 'DNSTRate': '', 'DNSTRate.Rank': '', '3PARate': '', '3PARate.Rank': '', 'D3PARate': '', 'D3PARate.Rank': '', 'ARate': '', 'ARate.Rank': '', 'DARate': '', 'DARate.Rank': '', 'PD3': '', 'PD3.Rank': '', 'DPD3': '', 'DPD3.Rank': '', 'PD2': '', 'PD2.Rank': '', 'DPD2': '', 'DPD2.Rank': '', 'PD1': '', 'PD1.Rank': '', 'DPD1': '', 'DPD1.Rank': ''}	
	for stat in stats:
		stats_df[stat[0]] = stat[1]
		stats_df[stat[0]+'.Rank'] = stat[2]
	return stats_df



import pytest
import kenpompy.summary as kpsum

def test_get_efficiency(browser):
	expected = ['Louisville', 'ACC', '67.2', '199', '68.3', '217', '17.6', '183', '17.5', '175', '113.7', '28', '107.6',
				'75', '94.4', '24', '98.8', '62']

	df = kpsum.get_efficiency(browser, season = '2019')
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected
	assert df.loc[44]['Team'] == 'Cal St. Northridge'

	expected = ['Louisville', 'BE', '65.3', '160', '67.1', '169', '113.1', '40', '107.3', '67', '87.7', '4', '91.2', 
				'7']

	df = kpsum.get_efficiency(browser, season = '2008')
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpsum.get_efficiency(browser, season = '1998')


def test_get_fourfactors(browser):
	expected = ['Louisville', 'ACC', '67.2', '199', '113.7', '28', '50.9', '165', '17.4', '106', '29.2', '137', '34.9', '115', 
				'94.4', '24', '46.8', '25', '16.1', '313', '25.9', '74', '32.0', '156']

	df = kpsum.get_fourfactors(browser, season = '2019')
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpsum.get_fourfactors(browser, season = '1998')


def test_get_teamstats(browser):
	expected = ['Louisville', 'ACC', '34.2', '177', '50.6', '161', '77.7', '8', '10.6', '272', '8.3', '101', '9.2', 
				'128', '53.6', '122', '43.7', '56', '113.7', '28']

	df = kpsum.get_teamstats(browser, season = '2019')
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected

	expected = ['Louisville', 'ACC', '32.0', '53', '46.0', '25', '69.5', '107', '7.9', '234', '6.3', '340', '9.8', 
				'153', '47.9', '70', '37.2', '129', '94.4', '24']

	df = kpsum.get_teamstats(browser, season = '2019', defense = True)
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpsum.get_teamstats(browser, season = '1998')


def test_get_pointdist(browser):
	expected = ['Louisville', 'ACC', '21.1', '60', '44.2', '311', '34.8', '105', '19.2', '137', '49.9', '161', '30.9', 
				'211']

	df = kpsum.get_pointdist(browser, season = '2019')
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpsum.get_pointdist(browser, season = '1998')


def test_get_height(browser):
	expected = ['Louisville', 'ACC', '76.8', '181', '+0.6', '87', '+1.3', '44', '-0.1', '190', '-0.1', '189', '-0.9', 
				'272', '-0.5', '233', '2.18', '29', '33.3', '97', '36.7', '273']

	df = kpsum.get_height(browser, season = '2019')
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected

	expected = ['Louisville', 'BE', '77.1', '90', '+2.2', '56', '+1.3', '49', '+0.8', '83', '-0.6', '225', '-0.8', 
				'226', '+0.7', '99', '1.45', '161', '34.8', '66']

	df = kpsum.get_height(browser, season = '2007')
	assert [str(i) for i in df[df.Team == 'Louisville'].iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpsum.get_height(browser, season = '2006')


def test_get_playerstats(browser):
	expected = ['35', 'Montrezl Harrell', 'Louisville', '61.2', '6-8', '235', 'So']

	df = kpsum.get_playerstats(browser, season = '2014')
	assert [str(i) for i in df[df.Player == 'Montrezl Harrell'].iloc[0].to_list()] == expected

	expected = ['2', 'Montrezl Harrell', 'Louisville', '61.2', '6-8', '235', 'So']

	df = kpsum.get_playerstats(browser, season = '2014', conf = "AMER")
	assert [str(i) for i in df[df.Player == 'Montrezl Harrell'].iloc[0].to_list()] == expected

	expected = ['2', 'Montrezl Harrell', 'Louisville', '59.8', '6-8', '235', 'So']

	df = kpsum.get_playerstats(browser, season = '2014', conf = "AMER", conf_only = True)
	assert [str(i) for i in df[df.Player == 'Montrezl Harrell'].iloc[0].to_list()] == expected

	expected = ['22', 'Russ Smith', 'Louisville', '108.9', '6-0', '165', 'Jr', '32.0']

	dfs = kpsum.get_playerstats(browser, season = '2013', metric = "ORTG")
	df = dfs[0]
	assert [str(i) for i in df[df.Player == 'Russ Smith'].iloc[0].to_list()] == expected


	with pytest.raises(ValueError):
		kpsum.get_playerstats(browser, season = '2003')

	with pytest.raises(ValueError):
		kpsum.get_playerstats(browser, season = '2012', conf_only = True)

	with pytest.raises(KeyError):
		kpsum.get_playerstats(browser, metric = "yeet")


def test_get_kpoy(browser):
	expected = ['1', 'Russ Smith', '2.636', '165', 'Jr', 'Briarwood, NY', 'Louisville ', '6-0']

	dfs = kpsum.get_kpoy(browser, season = '2013')
	df = dfs[0]
	assert [str(i) for i in df[df.Player == 'Russ Smith'].iloc[0].to_list()] == expected

	expected = ['10.0', 'Russ Smith', '15', '165', 'Jr', 'Briarwood, NY', 'Louisville ', '6-0']

	df = dfs[1]
	assert [str(i) for i in df[df.Player == 'Russ Smith'].iloc[0].to_list()] == expected

	dfs = kpsum.get_kpoy(browser, season = '2011')
	df = dfs[0]
	assert len(dfs) == 1

	with pytest.raises(ValueError):
		kpsum.get_kpoy(browser, season = '2010')



from kenpompy.FanMatch import FanMatch

def test_fanmatch(browser):
	date = "2020-01-29"

	expected = ["31 Marquette 84, 59 Xavier 82 (2OT)",
				"Cincinnati, OH Cintas Center",
				"68.2",
				"8",
				"2.94",
				"2",
				"nan",
				"nan",
				"Sacar Anim (28p/5r/2a/1b/4s)",
				"nan",
				"79",
				"Marquette",
				"73-72",
				"51%",
                "nan",
				"1",
				"Xavier",
				"2OT",
				"Xavier",
				"59",
				"82",
				"Marquette",
				"31",
				"84",
				"2"]

	fm = FanMatch(browser, date)
	assert fm.fm_df.iloc[1,].astype("str").tolist() == expected
	assert fm.date == "2020-01-29"
	assert fm.url == 'https://kenpom.com/fanmatch.php?d=2020-01-29'
	assert fm.lines_o_night == ["1. Max Mahoney, Boston University • 29 pts (11-15 2's, 7-11 FT's) • 10 Rebs (5 Off) • 4 Assists • 3 Stls",
								"2. Jordan Nwora, Louisville • 37 pts (4-6 2's, 7-14 3's, 8-9 FT's) • 9 Rebs • 1 Stl • 1 Assist",
								"3. Maceo Jack, George Washington • 35 pts (5-6 2's, 7-13 3's, 4-6 FT's) • 7 Rebs • 2 Blks • 1 Stl"]
	assert fm.ppg == 71.5
	assert fm.avg_eff == 101.9
	assert fm.pos_40 == 68.7
	assert fm.mean_abs_err_pred_total_score == 15.8
	assert fm.bias_pred_total_score == -1.6
	assert fm.mean_abs_err_pred_mov == 7.8
	assert fm.record_favs == "40-13"
	assert fm.expected_record_favs == "38-15"
	assert fm.exact_mov == "1/53"

	# Tests for tournament label parsing (j-andrews7/kenpompy#47)
	date = "2023-03-03"
	fm = FanMatch(browser, date)
	assert "MVC-T" not in fm.fm_df.Game.loc[3]
	assert fm.fm_df.Tournament.loc[6] == "WCC-T"

	date = "2016-04-02"
	fm = FanMatch(browser, date)
	assert fm.fm_df.Tournament.loc[0] == "NCAA"
	assert "NCAA" not in fm.fm_df.Game.loc[0]

    # Test for a day with no game that is in season
	date = "2020-12-24"
	fm = FanMatch(browser, date)
	assert fm.fm_df is None

    # Test for a day with no games that is during the offseason
	date = "2024-10-30"
	fm = FanMatch(browser, date)
	assert fm.fm_df is None



import pytest
import kenpompy.misc as kpmisc

def test_get_current_season(browser):
	current_season = kpmisc.get_current_season(browser)

	# Test that the current season is indeed published on the homepage
	assert browser.request(url='https://kenpom.com/?y=' + str(current_season), method='GET', allow_redirects=False).status_code == 200

	# Test that there are isn't a season beyond the "current" one (that it is indeed the latest)
	assert browser.request(url='https://kenpom.com/?y=' + str(current_season + 1), method='GET', allow_redirects=False).status_code == 302

def test_get_pomeroy_ratings(browser):
    expected = ['1', 'Virginia', 'ACC', '35-3', '+34.22', '123.4', '2', '89.2', '5', '59.4', '353', '+.050', '62', '+11.18', '22', '109.2', '34', '98.1', '14', '-3.24', '255', '1']
    df = kpmisc.get_pomeroy_ratings(browser, season=2019)
    assert df.iloc[0].to_list() == expected

    expected = ['253', 'Cal St. Northridge', 'BW', '13-21', '-7.74', '104.8', '170', '112.5', '318', '70.9', '43', '-.018', '232', '-4.63', '263', '101.7', '282', '106.3', '225', '-4.09', '279', '']
    assert df.loc[252].to_list() == expected

    # Also test proper handling of team names with special characters like ' and &
    expected = ['31', "Saint Mary's", 'WCC', '22-12', '+17.31', '114.7', '23', '97.4', '55', '62.7', '348', '-.045', '285', '+3.66', '82', '106.6', '76', '103.0', '100', '-0.90', '183', '11']
    assert df.iloc[30].to_list() == expected

	# Shape test to ensure header rows are correctly filtered
    expected = (353, 22)
    assert df.shape == expected

def test_get_trends(browser):
	expected = ["2019","103.2","69.0","50.7","18.5","28.4","33.0","50.1","34.4","38.7","70.7","51.9","9.3","8.9","9.7",
				"76.8","47.8","59.0","71.9"]

	df = kpmisc.get_trends(browser)
	assert [str(i) for i in df[df.Season == "2019"].iloc[0].to_list()] == expected


def test_get_refs(browser):
	expected = ["2",
				"Keith Kimble",
				"67.08",
				"107",
				"Sat 4/6",
				"1 Virginia 63, 11 Auburn 62 (Minneapolis, MN)"]
	df = kpmisc.get_refs(browser, season = "2019")
	assert [str(i) for i in df[df.Name == "Keith Kimble"].iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpmisc.get_refs(browser, season = "2014")


def test_get_hca(browser):
	expected = len(['Louisville', 'ACC', '3.4', '119', '-4.1','44', '4.5', '269',
		'-2.1', '14', '1.1', '228', '400', '185'])

	df = kpmisc.get_hca(browser)
	assert len(df[df.Team == "Louisville"].iloc[0].to_list()) == expected


def test_get_arenas(browser):
	expected = ["3", "Louisville", "ACC", "KFC Yum! Center", "nan", "22,000", "nan"]

	df = kpmisc.get_arenas(browser, season = "2019")
	assert [str(i) for i in df[df.Team == "Louisville"].iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpmisc.get_arenas(browser, season = "2009")


def test_get_gameattribs(browser):
	expected = ["1",
				"Tue Feb 26",
				"126 West Virginia 104, 42 TCU 96 (3OT)",
				"Morgantown, WV",
				"B12",
				"4.08",
				"WVU Coliseum"]
	df = kpmisc.get_gameattribs(browser, season = "2019")
	assert [str(i) for i in df.iloc[0].to_list()] == expected

	with pytest.raises(ValueError):
		kpmisc.get_gameattribs(browser, season = "2009")

	with pytest.raises(KeyError):
		kpmisc.get_gameattribs(browser, season = "2019", metric = "yeet")

	with pytest.raises(ValueError):
		kpmisc.get_gameattribs(browser, season = "2010", metric = "BUSTS")


def test_get_program_ratings(browser):
	df = kpmisc.get_program_ratings(browser)
	expected = (364, 17)
	assert df.shape[1] == expected[1]


from sys import version_info
from requests.adapters import HTTPAdapter
from urllib3.util.ssl_ import create_urllib3_context

CIPHERS = (
    ':@SECLEVEL=2:ECDH+AESGCM:ECDH+CHACHA20:ECDH+AES:DHE+AES:!aNULL:!eNULL:!aDSS:!SHA1:!AESCCM'
)

def environment_requires_DES_adapter():
    return version_info.major == 3 and version_info.minor < 11

class DESAdapter(HTTPAdapter):
    """
    A TransportAdapter that re-enables 3DES support in Requests to avoid Cloudflare filtering based on SSL profiling
    Adapted from the research provided by Nick Ostendorf (@nickostendorf) in https://github.com/j-andrews7/kenpompy/issues/33
    """

    def init_poolmanager(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs['ssl_context'] = context
        return super(DESAdapter, self).init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs['ssl_context'] = context
        return super(DESAdapter, self).proxy_manager_for(*args, **kwargs)



import kenpompy.conference as kpconf

def test_get_valid_conferences(browser):
	expected = 32

	confs_2021 = kpconf.get_valid_conferences(browser, season = '2021')
	assert len(confs_2021) == expected

	valid_2021_confs = ['B12', 'P12', 'Pat']
	for team in valid_2021_confs:
		assert team in confs_2021
	assert confs_2021[0] == 'A10'
	assert confs_2021[-1] == 'WCC'

	confs_2003 = kpconf.get_valid_conferences(browser, season = '2003')
	assert len(confs_2003) == 32

	valid_2003_confs = ['MEAC', 'CAA']
	for team in valid_2003_confs:
		assert team in confs_2003
	assert confs_2003[0] == 'A10'
	assert confs_2003[-1] == 'ind'


def test_get_aggregate_stats(browser):
	expectedTempo = 67.9
	expectedTempoRank = 20
	expectedHomeWinPercent = 60.3
	expectedHomeWinPercentRank = 8

	confs_2021 = kpconf.get_aggregate_stats(browser, 'B10', season = '2021')
	assert confs_2021.loc['Tempo', 'Value'] == expectedTempo
	assert confs_2021.loc['Tempo', 'Rank'] == expectedTempoRank
	assert confs_2021.loc['Home win%', 'Value'] == expectedHomeWinPercent
	assert confs_2021.loc['Home win%', 'Rank'] == expectedHomeWinPercentRank
	
	expectedTempo = 65.5
	expectedTempoRank = 27
	expectedHomeWinPercent = 71.6
	expectedHomeWinPercentRank = 2

	confs_2003 = kpconf.get_aggregate_stats(browser, 'B10', season = '2003')
	assert confs_2003.loc['Tempo', 'Value'] == expectedTempo
	assert confs_2003.loc['Tempo', 'Rank'] == expectedTempoRank
	assert confs_2003.loc['Home win%', 'Value'] == expectedHomeWinPercent
	assert confs_2003.loc['Home win%', 'Rank'] == expectedHomeWinPercentRank

	expectedTempo = 67
	expectedTempoRank = 28

	confs_2021 = kpconf.get_aggregate_stats(browser, season = '2021')
	assert confs_2021.loc['AE', 'Tempo'] == expectedTempo
	assert confs_2021.loc['AE', 'Tempo.Rank'] == expectedTempoRank

	expectedTempo = 67.2
	expectedTempoRank = 20

	confs_2003 = kpconf.get_aggregate_stats(browser, season = '2003')
	assert confs_2003.loc['AE', 'Tempo'] == expectedTempo
	assert confs_2003.loc['AE', 'Tempo.Rank'] == expectedTempoRank


def test_get_standings(browser):
	expectedTeam1 = 'Michigan'
	expectedTeam1AdjO = 117.6
	expectedTeam1AdjORank = 9
	
	confs_2021 = kpconf.get_standings(browser, 'B10', season = '2021')
	assert confs_2021.iloc[0, :]['Team'] == expectedTeam1
	assert confs_2021.iloc[0, :]['ORtg'] == expectedTeam1AdjO
	assert confs_2021.iloc[0, :]['ORtg.Rank'] == expectedTeam1AdjORank

	expectedTeam1 = 'Wisconsin'
	expectedTeam1AdjO = 114.4
	expectedTeam1AdjORank = 16	

	confs_2003 = kpconf.get_standings(browser, 'B10', season = '2003')
	assert confs_2003.iloc[0, :]['Team'] == expectedTeam1
	assert confs_2003.iloc[0, :]['ORtg'] == expectedTeam1AdjO
	assert confs_2003.iloc[0, :]['ORtg.Rank'] == expectedTeam1AdjORank


def test_get_offense(browser):
	expectedTeam1 = 'Iowa'
	expectedTeam1Tempo = 69
	expectedTeam1TempoRank = 4
	
	confs_2021 = kpconf.get_offense(browser, 'B10', season = '2021')
	assert confs_2021.iloc[0, :]['Team'] == expectedTeam1
	assert confs_2021.iloc[0, :]['Tempo'] == expectedTeam1Tempo
	assert confs_2021.iloc[0, :]['Tempo.Rank'] == expectedTeam1TempoRank

	expectedTeam1 = 'Wisconsin'
	expectedTeam1Tempo = 61.5
	expectedTeam1TempoRank = 11	

	confs_2003 = kpconf.get_offense(browser, 'B10', season = '2003')
	assert confs_2003.iloc[0, :]['Team'] == expectedTeam1
	assert confs_2003.iloc[0, :]['Tempo'] == expectedTeam1Tempo
	assert confs_2003.iloc[0, :]['Tempo.Rank'] == expectedTeam1TempoRank


def test_get_defense(browser):
	expectedTeam1 = 'Michigan'
	expectedTeam1Stl = 6.4
	expectedTeam1StlRank = 12
	
	confs_2021 = kpconf.get_defense(browser, 'B10', season = '2021')
	assert confs_2021.iloc[0, :]['Team'] == expectedTeam1
	assert confs_2021.iloc[0, :]['Stl%'] == expectedTeam1Stl
	assert confs_2021.iloc[0, :]['Stl%.Rank'] == expectedTeam1StlRank

	expectedTeam1 = 'Illinois'
	expectedTeam1Stl = 9.7
	expectedTeam1StlRank = 6

	confs_2003 = kpconf.get_defense(browser, 'B10', season = '2003')
	assert confs_2003.iloc[0, :]['Team'] == expectedTeam1
	assert confs_2003.iloc[0, :]['Stl%'] == expectedTeam1Stl
	assert confs_2003.iloc[0, :]['Stl%.Rank'] == expectedTeam1StlRank

	confs_2021 = kpconf.get_defense(browser, 'BW', season = '2021')
	assert confs_2021.loc[5]['Team'] == 'Cal St. Bakersfield'


"""
This module provides functions for scraping the summary stats kenpom.com pages into more
usable pandas dataframes.
"""

import pandas as pd
import re
from io import StringIO
from cloudscraper import CloudScraper
from bs4 import BeautifulSoup
from typing import Optional
from .utils import get_html

def get_efficiency(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the Efficiency stats table (https://kenpom.com/summary.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 1999 is the earliest available season but 
			possession length data wasn't available until 2010. Most recent season is the default.

	Returns:
		eff_df (pandas dataframe): Pandas dataframe containing the summary efficiency/tempo table from kenpom.com.

	Raises:
		ValueError: If `season` is less than 1999.
	"""

	url = 'https://kenpom.com/summary.php'

	if season:
		if int(season) < 1999:
			raise ValueError(
				'season cannot be less than 1999, as data only goes back that far.')
		url = url + '?y=' + str(season)

	eff = BeautifulSoup(get_html(browser, url), "html.parser")
	table = eff.find_all('table')[0]
	eff_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	eff_df = eff_df[0]

	# Handle seasons prior to 2010 having fewer columns.
	if len(eff_df.columns) == 18:
		eff_df = eff_df.iloc[:, 0:18]
		eff_df.columns = ['Team', 'Conference', 'Tempo-Adj', 'Tempo-Adj.Rank', 'Tempo-Raw', 'Tempo-Raw.Rank',
						  'Avg. Poss Length-Offense', 'Avg. Poss Length-Offense.Rank', 'Avg. Poss Length-Defense',
						  'Avg. Poss Length-Defense.Rank', 'Off. Efficiency-Adj', 'Off. Efficiency-Adj.Rank',
						  'Off. Efficiency-Raw', 'Off. Efficiency-Raw.Rank', 'Def. Efficiency-Adj',
						  'Def. Efficiency-Adj.Rank', 'Def. Efficiency-Raw', 'Def. Efficiency-Raw.Rank']
	else:
		eff_df = eff_df.iloc[:, 0:14]
		eff_df.columns = ['Team', 'Conference', 'Tempo-Adj', 'Tempo-Adj.Rank', 'Tempo-Raw', 'Tempo-Raw.Rank',
						  'Off. Efficiency-Adj', 'Off. Efficiency-Adj.Rank', 'Off. Efficiency-Raw',
						  'Off. Efficiency-Raw.Rank', 'Def. Efficiency-Adj', 'Def. Efficiency-Adj.Rank',
						  'Def. Efficiency-Raw', 'Def. Efficiency-Raw.Rank']

	# Remove the header rows that are interjected for readability.
	eff_df = eff_df[eff_df.Team != 'Team']
	# Remove NCAA tourny seeds for previous seasons.
	eff_df['Team'] = eff_df['Team'].str.replace(r'\d+', '', regex=True)
	eff_df['Team'] = eff_df['Team'].str.rstrip()
	eff_df = eff_df.dropna()

	return eff_df


def get_fourfactors(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the Four Factors table (https://kenpom.com/stats.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.
			Most recent season is the default.

	Returns:
		ff_df (pandas dataframe): Pandas dataframe containing the summary Four Factors table from kenpom.com.

	Raises:
		ValueError: If `season` is less than 1999.
	"""

	url = 'https://kenpom.com/stats.php'

	if season:
		if int(season) < 1999:
			raise ValueError(
				'season cannot be less than 1999, as data only goes back that far.')
		url = url + '?y=' + str(season)

	ff = BeautifulSoup(get_html(browser, url), "html.parser")
	table = ff.find_all('table')[0]
	ff_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	ff_df = ff_df[0]
	ff_df = ff_df.iloc[:, 0:24]
	ff_df.columns = ['Team', 'Conference', 'AdjTempo', 'AdjTempo.Rank', 'AdjOE', 'AdjOE.Rank', 'Off-eFG%',
					 'Off-eFG%.Rank', 'Off-TO%', 'Off-TO%.Rank', 'Off-OR%', 'Off-OR%.Rank', 'Off-FTRate',
					 'Off-FTRate.Rank', 'AdjDE', 'AdjDE.Rank', 'Def-eFG%', 'Def-eFG%.Rank', 'Def-TO%', 'Def-TO%.Rank',
					 'Def-OR%', 'Def-OR%.Rank', 'Def-FTRate', 'Def-FTRate.Rank']

	# Remove the header rows that are interjected for readability.
	ff_df = ff_df[ff_df.Team != 'Team']
	# Remove NCAA tourny seeds for previous seasons.
	ff_df['Team'] = ff_df['Team'].str.replace(r'\d+', '', regex=True)
	ff_df['Team'] = ff_df['Team'].str.rstrip()
	ff_df = ff_df.dropna()

	return ff_df


def get_teamstats(browser: CloudScraper, defense: Optional[bool]=False, season: Optional[str]=None):
	"""
	Scrapes the Miscellaneous Team Stats table (https://kenpom.com/teamstats.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		defense (bool, optional): Used to flag whether the defensive teamstats table is wanted or not. False by 
			default.
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.
			Most recent season is the default.

	Returns:
			ts_df (pandas dataframe): Pandas dataframe containing the Miscellaneous Team Stats table from kenpom.com.

	Raises:
			ValueError: If `season` is less than 1999.
	"""

	url = 'https://kenpom.com/teamstats.php'
	last_cols = ['AdjOE', 'AdjOE.Rank']

	# Create URL.
	if season:
		if int(season) < 1999:
			raise ValueError(
				'season cannot be less than 1999, as data only goes back that far.')
		url = url + '?y=' + str(season)
		if defense:
			url = url + '&od=d'
			last_cols = ['AdjDE', 'AdjDE.Rank']
	elif defense:
		url = url + '?od=d'
		last_cols = ['AdjDE', 'AdjDE.Rank']

	ts = BeautifulSoup(get_html(browser, url), "html.parser")
	table = ts.find_all('table')[0]
	ts_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	ts_df = ts_df[0]
	ts_df = ts_df.iloc[:, 0:20]
	ts_df.columns = ['Team', 'Conference', '3P%', '3P%.Rank', '2P%', '2P%.Rank', 'FT%', 'FT%.Rank',
					 'Blk%', 'Blk%.Rank', 'Stl%', 'Stl%.Rank', 'NST%', 'NST%.Rank', 'A%', 'A%.Rank', '3PA%', '3PA%.Rank',
					 last_cols[0], last_cols[1]]

	# Remove the header rows that are interjected for readability.
	ts_df = ts_df[ts_df.Team != 'Team']
	# Remove NCAA tourny seeds for previous seasons.
	ts_df['Team'] = ts_df['Team'].str.replace(r'\d+', '', regex=True)
	ts_df['Team'] = ts_df['Team'].str.rstrip()
	ts_df = ts_df.dropna()

	return ts_df


def get_pointdist(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the Team Points Distribution table (https://kenpom.com/pointdist.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.
			Most recent season is the default.

	Returns:
		dist_df (pandas dataframe): Pandas dataframe containing the Team Points Distribution table from kenpom.com.

	Raises:
		ValueError: If `season` is less than 1999.
	"""

	url = 'https://kenpom.com/pointdist.php'

	# Create URL.
	if season:
		if int(season) < 1999:
			raise ValueError(
				'season cannot be less than 1999, as data only goes back that far.')
		url = url + '?y=' + str(season)

	dist = BeautifulSoup(get_html(browser, url), "html.parser")
	table = dist.find_all('table')[0]
	dist_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	dist_df = dist_df[0]
	dist_df = dist_df.iloc[:, 0:14]
	dist_df.columns = ['Team', 'Conference', 'Off-FT', 'Off-FT.Rank', 'Off-2P', 'Off-2P.Rank', 'Off-3P', 'Off-3P.Rank',
					   'Def-FT', 'Def-FT.Rank', 'Def-2P', 'Def-2P.Rank', 'Def-3P', 'Def-3P.Rank']

	# Remove the header rows that are interjected for readability.
	dist_df = dist_df[dist_df.Team != 'Team']
	# Remove NCAA tourny seeds for previous seasons.
	dist_df['Team'] = dist_df['Team'].str.replace(r'\d+', '', regex=True)
	dist_df['Team'] = dist_df['Team'].str.rstrip()
	dist_df = dist_df.dropna()

	return dist_df


def get_height(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the Height/Experience table (https://kenpom.com/height.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 2007 is the earliest available season but 
			continuity data wasn't available until 2008. Most recent season is the default.

	Returns:
		h_df (pandas dataframe): Pandas dataframe containing the Height/Experience table from kenpom.com.

	Raises:
		ValueError: If `season` is less than 2007.
	"""

	url = 'https://kenpom.com/height.php'

	if season:
		if int(season) < 2007:
			raise ValueError(
				'Season cannot be less than 2007, as data only goes back that far.')
		url = url + '?y=' + str(season)

	height = BeautifulSoup(get_html(browser, url), "html.parser")
	table = height.find_all('table')[0]
	h_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	h_df = h_df[0]

	# Handle seasons prior to 2008 having fewer columns.
	if len(h_df.columns) == 22:
		h_df = h_df.iloc[:, 0:22]
		h_df.columns = ['Team', 'Conference', 'AvgHgt', 'AvgHgt.Rank', 'EffHgt', 'EffHgt.Rank',
						'C-Hgt', 'C-Hgt.Rank', 'PF-Hgt', 'PF-Hgt.Rank', 'SF-Hgt', 'SF-Hgt.Rank',
						'SG-Hgt', 'SG-Hgt.Rank', 'PG-Hgt', 'PG-Hgt.Rank', 'Experience', 'Experience.Rank',
						'Bench', 'Bench.Rank', "Continuity", "Continuity.Rank"]
	else:
		h_df = h_df.iloc[:, 0:20]
		h_df.columns = ['Team', 'Conference', 'AvgHgt', 'AvgHgt.Rank', 'EffHgt', 'EffHgt.Rank',
						'C-Hgt', 'C-Hgt.Rank', 'PF-Hgt', 'PF-Hgt.Rank', 'SF-Hgt', 'SF-Hgt.Rank',
						'SG-Hgt', 'SG-Hgt.Rank', 'PG-Hgt', 'PG-Hgt.Rank', 'Experience', 'Experience.Rank',
						'Bench', 'Bench.Rank']

	# Remove the header rows that are interjected for readability.
	h_df = h_df[h_df.Team != 'Team']
	# Remove NCAA tourny seeds for previous seasons.
	h_df['Team'] = h_df['Team'].str.replace(r'\d+', '', regex=True)
	h_df['Team'] = h_df['Team'].str.rstrip()
	h_df = h_df.dropna()

	return h_df


def get_playerstats(browser: CloudScraper, season: Optional[str]=None, metric: str='EFG', conf: Optional[str]=None, conf_only: bool=False):
	"""
	Scrapes the Player Leaders tables (https://kenpom.com/playerstats.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 2004 is the earliest available season. 
			Most recent season is the default.
		metric (str, optional): Used to get leaders for different metrics. Available values are: 'ORtg', 'Min', 
			'eFG', 'Poss', 'Shots', 'OR', 'DR', 'TO', 'ARate', 'Blk', 'FTRate', 'Stl', 'TS', 'FC40', 'FD40', '2P', 
			'3P', 'FT'. Default is 'eFG'. 'ORtg' returns a list of four dataframes, as there are four tables: 
			players that used >28% of possessions, >24% of possessions, >20% of possessions, and with no possession 
			restriction.
		conf (str, optional): Used to limit to players in a specific conference. Allowed values are: 'A10', 'ACC',
			'AE', 'AMER', 'ASUN', 'B10', 'B12', 'BE', 'BSKY', 'BSTH', 'BW', 'CAA', 'CUSA', 'HORZ', 'IND', IVY', 
			'MAAC', 'MAC', 'MEAC', 'MVC', 'MWC', 'NEC', 'OVC', 'P12', 'PAT', 'SB', 'SC', 'SEC', 'SLND', 'SUM', 
			'SWAC', 'WAC', 'WCC'. If you try to use a conference that doesn't exist for a given season, like 'IND'
			and '2018', you'll get an empty table, as kenpom.com doesn't serve 404 pages for invalid table queries
			like that. No filter applied by default.
		conf_only (bool, optional): Used to define whether stats should reflect conference games only. Only
			available if specific conference is defined. Only available for seasons after 2013. False by default.


	Returns:
		ps_df (pandas dataframe): Pandas dataframe containing the Player Leaders table from kenpom.com.

	Raises:
		ValueError: If `season` is less than 2004 or `conf_only` is used with an invalid `season`.
		KeyError: If `metric` is invalid.
	"""

	# `metric` parameter checking.
	metric = metric.upper()
	metrics = {'ORTG': 'ORtg', 'MIN': 'PctMin', 'EFG': 'eFG', 'POSS': 'PctPoss', 'SHOTS': 'PctShots', 'OR': 'ORPct', 
			   'DR': 'DRPct', 'TO': 'TORate', 'ARATE': 'ARate', 'BLK': 'PctBlocks', 'FTRATE': 'FTRate', 
			   'STL': 'PctStls', 'TS': 'TS', 'FC40': 'FCper40', 'FD40': 'FDper40', '2P': 'FG2Pct', '3P': 'FG3Pct', 
			   'FT': 'FTPct'}
	if metric not in metrics:
		raise KeyError(
			"""Metric is invalid, must be one of: 'ORtg', 'Min', 'eFG', 'Poss', 'Shots', 'OR', 'DR', 'TO', 'ARate', 
			'Blk', 'FTRate', 'Stl', 'TS', 'FC40', 'FD40', '2P', '3P', 'FT'""")
	else:
		met_url = 's=' + metrics[metric]


	url = 'https://kenpom.com/playerstats.php?' + met_url

	if season:
		if int(season) < 2004:
			raise ValueError(
				'Season cannot be less than 2004, as data only goes back that far.')
		elif int(season) < 2014 and conf_only:
			raise ValueError(
				'Conference only stats only available for seasons after 2013.'
			)
		url = url + '&y=' + str(season)

	if conf_only:
		url = url + '&c=c'

	if conf:
		url = url + '&f=' + conf

	playerstats = BeautifulSoup(get_html(browser, url), "html.parser")
	if metric == 'ORTG':
		ps_dfs = []
		tables = playerstats.find_all('table')
		for t in tables:
			ps_df = pd.read_html(StringIO(str(t)))
			ps_df = ps_df[0]
			
			# Split ortg column.
			ps_df.columns = ['Rank', 'Player', 'Team', 'ORtg', 'Ht', 'Wt', 'Yr']
			ps_df[['ORtg', 'Poss%']] = ps_df['ORtg'].str.split(' ', n=1, expand=True)
			ps_df['Poss%'] = ps_df['Poss%'].str.strip('()')

			ps_df = ps_df[ps_df.Rank != 'Rk']
			ps_df = ps_df.dropna()

			ps_dfs.append(ps_df)
		ps_df = ps_dfs
	else:
		perc_mets = ['Min', 'eFG', 'Poss', 'Shots', 'OR', 'DR', 'TO', 'Blk', 'Stl', 'TS', '2P', '3P', 'FT']
		if metric.upper() in perc_mets:
			metric = metric + '%'
		table = playerstats.find_all('table')[0]
		ps_df = pd.read_html(StringIO(str(table)))

		# Dataframe tidying.
		ps_df = ps_df[0]

		if metric.upper() in ['2P', '3P', 'FT']:
			ps_df.columns = ['Rank', 'Player', 'Team', metric.rstrip('%') + 'M', 
			metric.rstrip('%') + 'A', metric, 'Ht', 'Wt', 'Yr'] 
		else:
			ps_df.columns = ['Rank', 'Player', 'Team', metric, 'Ht', 'Wt', 'Yr'] 

		# Remove the header rows that are interjected for readability.
		ps_df = ps_df[ps_df.Rank != 'Rk']
		ps_df = ps_df.dropna()

	return ps_df


def get_kpoy(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the kenpom Player of the Year tables (https://kenpom.com/kpoy.php) into dataframes.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 2011 is the earliest available season.
			Most recent season is the default.

	Returns:
		kpoy_dfs (list of pandas dataframe): List of dandas dataframes containing the kenpom Player of the Year
			and Game MVP leaders tables from kenpom.com. Game MVP table only available from 2013 season onwards.

	Raises:
		ValueError: If `season` is less than 2011.
	"""

	kpoy_dfs = []
	url = 'https://kenpom.com/kpoy.php'

	# Create URL.
	if season:
		if int(season) < 2011:
			raise ValueError(
				'season cannot be less than 2011, as data only goes back that far.')
		url = url + '?y=' + str(season)
	else:
		season = 2013

	kpoy = BeautifulSoup(get_html(browser, url), "html.parser")
	table = kpoy.find_all('table')[0]
	df = pd.read_html(StringIO(str(table)))

	kpoy_df = df[0]
	kpoy_df.columns = ['Rank', 'Player', 'KPOY Rating']

	# Some mildly moronic dataframe tidying.
	kpoy_df[['Player', 'Weight', 'Year', 'Hometown']] = kpoy_df['Player'].str.split(' · ', expand=True)
	kpoy_df[['Player', 'Info']] = kpoy_df['Player'].str.split(', ', n=1, expand=True)
	kpoy_df['Team'] = kpoy_df['Info'].str.replace(r'\d+', '', regex=True).str.rstrip('-')
	kpoy_df['Height'] = kpoy_df['Info'].str.replace(r'[a-z]+', '', flags=re.IGNORECASE, regex=True).str.strip('. ').str.strip()
	kpoy_df = kpoy_df.drop(['Info'], axis=1)

	kpoy_dfs.append(kpoy_df)
	# Now the MVP table.
	if int(season) >= 2013:
		table = kpoy.find_all('table')[-1]
		df = pd.read_html(StringIO(str(table)))

		mvp_df = df[0]
		mvp_df.columns = ['Rank', 'Player', 'Game MVPs']

		# More tidying.
		mvp_df[['Player', 'Weight', 'Year', 'Hometown']] = mvp_df['Player'].str.split(' · ', expand=True)
		mvp_df[['Player', 'Info']] = mvp_df['Player'].str.split(', ', n=1, expand=True)
		mvp_df['Team'] = mvp_df['Info'].str.replace(r'\d+', '', regex=True).str.rstrip('-')
		mvp_df['Height'] = mvp_df['Info'].str.replace(r'[a-z]+', '', flags=re.IGNORECASE, regex=True).str.strip('. ').str.strip()
		mvp_df = mvp_df.drop(['Info'], axis=1)

		kpoy_dfs.append(mvp_df)

	return kpoy_dfs



"""
This module provides functions for scraping the miscellaneous stats kenpom.com pages into more
usable pandas dataframes.
"""

import pandas as pd
from io import StringIO
import re
from cloudscraper import CloudScraper
from bs4 import BeautifulSoup
from typing import Optional
from .utils import get_html

def get_current_season(browser: CloudScraper):
	"""
	Scrapes the KenPom homepage to get the latest season year that has data published

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
            by the `login` function.

	Returns:
		current_season (int): Number corresponding to the last season year that has data published
	"""
	url = 'https://kenpom.com/index.php'
	content = BeautifulSoup(get_html(browser, url), "html.parser")
	page_title = content.select_one('#content-header h2').text
	YEAR_PATTERN = r'^(\d{4})'
	return int(re.match(YEAR_PATTERN, page_title).group(0))

def get_pomeroy_ratings(browser: CloudScraper, season: Optional[str]=None):
    """
    Scrapes the Pomeroy College Basketball Ratings table (https://kenpom.com/index.php) into a dataframe.

    Args:
        browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
            by the `login` function.
        season (str, optional): Used to define different seasons. 1999 is the earliest available season.
            Most recent season is the default.
    Returns:
        refs_df (pandas dataframe): Pandas dataframe containing the Pomeroy College Basketball Ratings table from kenpom.com.
    Raises:
        ValueError: If `season` is less than 1999.
    """
    url = 'https://kenpom.com/index.php'
    if season and int(season) < 1999:
        raise ValueError("season cannot be less than 1999")
    url += '?y={}'.format(season)
    page = BeautifulSoup(get_html(browser, url), "html.parser")
    table = page.find_all('table')[0]
    ratings_df = pd.read_html(StringIO(str(table)))
    # Dataframe tidying.
    ratings_df = ratings_df[0]
    ratings_df.columns = ratings_df.columns.map(lambda x: x[1])
    ratings_df.dropna(inplace=True)
    ratings_df = ratings_df[ratings_df['Rk'] != 'Rk']
    ratings_df.reset_index(drop=True, inplace=True)
    # Parse out seed, most current won't have this
    tmp = ratings_df['Team'].str.extract(r'(?P<Team>[a-zA-Z.&\'\s]+(?<!\s))\s*(?P<Seed>\d*)')
    ratings_df["Team"] = tmp["Team"]
    ratings_df["Seed"] = tmp["Seed"]
    
    # Disambiguate column names for easier reference
    ratings_df.columns = ['Rk', 'Team', 'Conf', 'W-L', 'AdjEM', 'AdjO',
                          'AdjO.Rank', 'AdjD', 'AdjD.Rank', 'AdjT', 'AdjT.Rank',
						  'Luck', 'Luck.Rank', 'SOS-AdjEM', 'SOS-AdjEM.Rank', 'SOS-OppO', 'SOS-OppO.Rank',
						  'SOS-OppD', 'SOS-OppD.Rank', 'NCSOS-AdjEM', 'NCSOS-AdjEM.Rank', 'Seed']
    
    return ratings_df


def get_trends(browser: CloudScraper):
	"""
	Scrapes the statistical trends table (https://kenpom.com/trends.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.

	Returns:
		trends_df (pandas dataframe): Pandas dataframe containing the statistical trends table from kenpom.com.
	"""

	url = 'https://kenpom.com/trends.php'

	trends = BeautifulSoup(get_html(browser, url), "html.parser")
	table = trends.find_all('table')[0]
	trends_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	trends_df = trends_df[0]
	trends_df.drop(trends_df.tail(5).index, inplace=True)

	return trends_df


def get_refs(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the officials rankings table (https://kenpom.com/officials.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 2016 is the earliest available season.
			Most recent season is the default.

	Returns:
		refs_df (pandas dataframe): Pandas dataframe containing the officials rankings table from kenpom.com.

	Raises:
		ValueError: If `season` is less than 2016.
	"""

	url = 'https://kenpom.com/officials.php'

	if season:
		if int(season) < 2016:
			raise ValueError(
				'season cannot be less than 2016, as data only goes back that far.')
		url = url + '?y=' + str(season)

	refs = BeautifulSoup(get_html(browser, url), "html.parser")
	table = refs.find_all('table')[0]
	refs_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	refs_df = refs_df[0]
	refs_df.columns = ['Rank', 'Name', 'Rating', 'Games', 'Last Game', 'Game Score', 'Box']
	refs_df = refs_df[refs_df.Rating != 'Rating']
	refs_df = refs_df.drop(['Box'], axis=1)

	return refs_df


def get_hca(browser: CloudScraper):
	"""
	Scrapes the home court advantage table (https://kenpom.com/hca.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.

	Returns:
		hca_df (pandas dataframe): Pandas dataframe containing the home court advantage table from kenpom.com.
	"""

	url = 'https://kenpom.com/hca.php'

	hca = BeautifulSoup(get_html(browser, url), "html.parser")
	table = hca.find_all('table')[0]
	hca_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	hca_df = hca_df[0]
	hca_df.columns = ['Team', 'Conference', 'HCA', 'HCA.Rank', 'PF', 'PF.Rank', 'Pts', 'Pts.Rank', 'NST',
						'NST.Rank', 'Blk', 'Blk.Rank', 'Elev', 'Elev.Rank']
	hca_df = hca_df[hca_df.Team != 'Team']

	return hca_df


def get_arenas(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the arenas table (https://kenpom.com/arenas.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 2010 is the earliest available season.
			Most recent season is the default.

	Returns:
		arenas_df (pandas dataframe): Pandas dataframe containing the arenas table from kenpom.com.

	Raises:
		ValueError: If `season` is less than 2010.
	"""

	url = 'https://kenpom.com/arenas.php'

	if season:
		if int(season) < 2010:
			raise ValueError(
				'season cannot be less than 2010, as data only goes back that far.')
		url = url + '?y=' + str(season)

	arenas = BeautifulSoup(get_html(browser, url), "html.parser")
	table = arenas.find_all('table')[0]
	arenas_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	arenas_df = arenas_df[0]
	arenas_df.columns = ['Rank', 'Team', 'Conference', 'Arena', 'Alternate']
	arenas_df[['Arena', 'Arena.Capacity']] = arenas_df['Arena'].str.split(r' \(', expand=True, regex=True)
	arenas_df['Arena.Capacity'] = arenas_df['Arena.Capacity'].str.rstrip(')')
	arenas_df[['Alternate', 'Alternate.Capacity']] = arenas_df['Alternate'].str.split(r' \(', expand=True, regex=True)
	arenas_df['Alternate.Capacity'] = arenas_df['Alternate.Capacity'].str.rstrip(')')

	return arenas_df


def get_gameattribs(browser: CloudScraper, season: Optional[str]=None, metric: str='Excitement'):
	"""
	Scrapes the Game Attributes tables (https://kenpom.com/game_attrs.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.
		season (str, optional): Used to define different seasons. 2010 is the earliest available season.
			Most recent season is the default.
		metric (str, optional): Used to get highest ranking games for different metrics. Available values are:
			'Excitement', 'Tension', 'Dominance', 'ComeBack', 'FanMatch', 'Upsets', and 'Busts'. Default is
			'Excitement'. 'FanMatch', 'Upsets', and 'Busts' are only valid for seasons after 2010.

	Returns:
		ga_df (pandas dataframe): Pandas dataframe containing the Game Attributes table from kenpom.com for a
		given metric.

	Raises:
		ValueError: If `season` is less than 2010.
		KeyError: If `metric` is invalid.
	"""

	# `metric` parameter checking.
	metric = metric.upper()
	metrics = {'EXCITEMENT': 'Excitement', 'TENSION': 'Tension', 'DOMINANCE': 'Dominance', 'COMEBACK': 'MinWP',
				'FANMATCH': 'FanMatch', 'UPSETS': 'Upsets', 'BUSTS': 'Busts'}
	if metric not in metrics:
		raise KeyError(
			"""Metric is invalid, must be one of: 'Excitement',
				'Tension', 'Dominance', 'ComeBack', 'FanMatch', 'Upsets', and 'Busts'""")
	else:
		met_url = 's=' + metrics[metric]

	url = 'https://kenpom.com/game_attrs.php?' + met_url

	# Season selection and an additional check.
	if season:
		if int(season) < 2010:
			raise ValueError(
				'Season cannot be less than 2010, as data only goes back that far.')
		elif int(season) < 2011 and metric.upper() in ['FANMATCH', 'UPSETS', 'BUSTS']:
			raise ValueError(
				'FanMatch, Upsets, and Busts tables only available for seasons after 2010.'
			)
		url = url + '&y=' + str(season)

	playerstats = BeautifulSoup(get_html(browser, url), "html.parser")

	table = playerstats.find_all('table')[0]
	ga_df = pd.read_html(StringIO(str(table)))

	# Dataframe tidying.
	ga_df = ga_df[0]
	ga_df.columns = ['Rank', 'Date', 'Game', 'Box', 'Location', 'Conf.Matchup', 'Value']
	ga_df = ga_df.drop(['Box'], axis=1)
	ga_df[['Location', 'Arena']] = ga_df['Location'].str.split(r' \(', expand=True, regex=True)
	ga_df['Arena'] = ga_df['Arena'].str.rstrip(')')

	return ga_df


def get_program_ratings(browser: CloudScraper):
	"""
	Scrapes the program ratings table (https://kenpom.com/programs.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function.

	Returns:
		programs_df (pandas dataframe): Pandas dataframe containing the program ratings table from kenpom.com.
	"""

	url = 'https://kenpom.com/programs.php'

	programs = BeautifulSoup(get_html(browser, url), "html.parser")
	table = programs.find_all('table')[0]
	programs_df = pd.read_html(StringIO(str(table)))
	programs_df = programs_df[0]

	programs_df.columns = ['Rank', 'Team', 'Conference', 'Rating', 'kenpom.Best.Rank', 'kenpom.Best.Season', 'kenpom.Worst.Rank',
							'kenpom.Worst.Season', 'kenpom.Median.Rank', 'kenpom.Top10.Finishes',
							'kenpom.Top25.Finishes', 'kenpom.Top50.Finishes', 'NCAA.Champs', 'NCAA.F4',
							'NCAA.S16', 'NCAA.R1', 'Change']

	programs_df = programs_df[programs_df.Team != 'Team']

	return programs_df



"""
The utils module provides utility functions, such as logging in.
"""

import cloudscraper
from cloudscraper import CloudScraper

def login(email: str, password: str):
	"""
	Logs in to kenpom.com using user credentials.

	Args:
		email (str): User e-mail for login to kenpom.com.
		password (str): User password for login to kenpom.com.

	Returns:
		browser (mechanicalsoup StatefulBrowser): Authenticated browser with full access to kenpom.com.
	"""

	browser = cloudscraper.create_scraper()
	browser.get('https://kenpom.com/index.php')

	form_data = {
		'email': email,
		'password': password,
		'submit': 'Login!',
	}

	# Response page actually throws an error but further navigation works and will show you as logged in.
	browser.post(
		'https://kenpom.com/handlers/login_handler.php',
		data=form_data, 
		allow_redirects=True
	)

	home_page = browser.get('https://kenpom.com/')
	if 'Logged in as' not in home_page.text:
		raise Exception('Logging in failed - check your credentials')

	return browser

def get_html(browser: CloudScraper, url: str):
	"""
	Performs a get request on the specified url.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
            by the `login` function.
		url (str): The url to perform the get request on.
	
	Returns:
		html (Bytes | Any): The return content.
	
	Raises:
		Exception if get request gets a non-200 response code.
	"""	
	response = browser.get(url)
	if response.status_code != 200:
		raise Exception(f'Failed to retrieve {url} (status code: {response.status_code})')
	return response.content


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kenpompy",
    version="0.5.0",
    author="Jared Andrews",
    author_email="jared.andrews07@gmail.com",
    description="A python package for scraping kenpom.com NCAA basketball data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/j-andrews7/kenpompy",
    packages=setuptools.find_packages(exclude=("tests", "docs")),
    license="GNU GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers"
    ],
    install_requires = ["mechanicalsoup", "pandas", "bs4", "cloudscraper"],
    python_requires='>=3.8',
)


"""
This module contains the FanMatch class for scraping the FanMatch pages into more usable objects.
"""

import pandas as pd
from io import StringIO
import re
from datetime import datetime
from cloudscraper import CloudScraper
from bs4 import BeautifulSoup
from typing import Optional
from .utils import get_html

class FanMatch:
    """Object to hold FanMatch page scraping results.
    
    This class scrapes the kenpom FanMatch page when a new instance is created. 

    Args:
        browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
            by the `login` function.
        date (str): Date to scrape, in format "YYYY-MM-DD", such as "2020-01-29".

    Attributes:
        url (str): Full url for the page to be scraped.
        date (str): Date to be scraped.
        lines_o_night (list): List containing lines of the night if games have taken place.
        ppg (float): Average points per game for the day.
        avg_eff (float): Average efficiency for the day.
        pos_40 (float): Possessions per 40 minutes for the day.
        mean_abs_err_pred_total_score (float): The mean absolute error of predicted total score for the day.
        bias_pred_total_score (float): The bias of predicted total score for the day.
        mean_abs_err_pred_mov (float): The mean absolute error of the predicted margin of victory for the day.
        record_favs (str): Record of favorites for the day.
        expected_record_favs (str): Expected record of favorites for the day.
        exact_mov (str): Number of games where margin of victory was accurately predicted out of total played.
        fm_df (pandas dataframe or None): Pandas dataframe containing parsed FanMatch table. If there are no games that day, fm_df will be None.
    """

    def __init__(self, browser: CloudScraper, date: Optional[str]=None):
        self.url = 'https://kenpom.com/fanmatch.php'
        self.date = date
        self.lines_o_night = None
        self.ppg = None
        self.avg_eff = None
        self.pos_40 = None
        self.mean_abs_err_pred_total_score = None
        self.bias_pred_total_score = None
        self.mean_abs_err_pred_mov = None
        self.record_favs = None
        self.expected_record_favs = None
        self.exact_mov = None
        self.fm_df = None
        
        if self.date is not None:
            self.url = self.url + "?d=" + self.date

        fm = BeautifulSoup(get_html(browser, self.url), "html.parser")
        if "Sorry, no games today." in fm.text:
            return
        if date is not None:
            date_text = fm.find("div", class_="lh12").get_text()
            date_match = re.search(r"for \w+, (\w+ \d{1,2}[a-z]{2})", date_text)
            if date_match:
                extracted_date_str = re.sub(r"(st|nd|rd|th)", "", date_match.group(1))
                extracted_date = datetime.strptime(extracted_date_str, "%B %d")
                extracted_mmdd = extracted_date.strftime("%m-%d")
                user_mmdd = datetime.strptime(date, "%Y-%m-%d").strftime("%m-%d")
                if extracted_mmdd != user_mmdd:
                    return
        table = fm.find_all("table")[0]
        fm_df = pd.read_html(StringIO(str(table)))
        fm_df = fm_df[0]
        fm_df = fm_df.rename(columns={"Thrill Score": "ThrillScore", "Come back": "Comeback", "Excite ment": "Excitement"})
        fm_df.ThrillScore = fm_df.ThrillScore.astype("str")
        fm_df["ThrillScoreRank"] = fm_df.ThrillScore.str[4:]
        fm_df["ThrillScoreRank"] = fm_df["ThrillScoreRank"].str.strip()
        fm_df.ThrillScore = fm_df.ThrillScore.str[0:4]
        
        # Take care of parsing if some/all games have been completed.
        if not all(pd.isnull(fm_df["Excitement"])):
            fm_df["Excitement"] = fm_df.Excitement.str.split("·").str[0]
            fm_df["ExcitementRank"] = fm_df.Excitement.str.split("·").str[1]
            
            # Handle extra rows without game info.
            e_start = fm_df.index[fm_df["Game"].str.contains("the night")].tolist()[0]
            extra = fm_df.iloc[e_start:len(fm_df),]
            fm_df.drop(extra.index, inplace = True)
            self.lines_o_night = extra.iloc[1:len(extra)-4, 0].tolist()
            
            sts = extra.iloc[-1, 0]
            pred_score = extra.iloc[-2, 0]
            pred_mov = extra.iloc[-3, 0]

            sts_s = sts.split(": ")[2:]
            sts_s = [x.split("•")[0].strip() for x in sts_s]
            self.ppg = float(sts_s[0])
            self.avg_eff = float(sts_s[1])
            self.pos_40  = float(sts_s[2])

            pred_s = pred_score.split(": ")[1:]
            pred_s = [x.split("•")[0].strip() for x in pred_s]
            self.mean_abs_err_pred_total_score = float(pred_s[0])
            self.bias_pred_total_score = float(pred_s[1])

            pred_m = pred_mov.split(": ")[1:]
            mean_abs_err_pred_mov = pred_m[0]
            self.mean_abs_err_pred_mov = float(mean_abs_err_pred_mov.split("  •")[0])
            record_favs = pred_m[1]
            self.record_favs = record_favs.split(" (")[0]
            expected_record_favs = pred_m[2]
            self.expected_record_favs = expected_record_favs.split(")")[0]
            exact_mov = pred_m[2]
            exact_mov = exact_mov.split(" in ")[1]
            exact_mov = exact_mov.split()
            self.exact_mov = exact_mov[0] + "/" + exact_mov[2]
    
        # Will only be present if some games have been completed.
        if not all(pd.isnull(fm_df.Comeback)):
            fm_df["Comeback"] = fm_df.Comeback.str.split("·").str[0]
            fm_df["ComebackRank"] = fm_df.Comeback.str.split("·").str[1]

        mvp = fm_df.Game.str.split(" MVP: ").str[1]
        fm_df["Game"], fm_df["MVP"] = fm_df.Game.str.split(" MVP: ").str[0], mvp

        # Conference tournament label handling (fixes j-andrews7/kenpompy#47)
        fm_df["Tournament"] = fm_df.Game.str.extract(r"([A-Za-z]{2,}-T|NCAA)$")
        fm_df["Game"] = fm_df.Game.str.replace(r"(\s+[A-Za-z]{2,}-T|NCAA)$", "", regex=True)

        pos = fm_df.Game.str.split(r" \[").str[1]
        fm_df["Game"], fm_df["Possessions"] = fm_df.Game.str.split(r" \[").str[0], pos.astype("str")
        fm_df.Possessions = fm_df.Possessions.str.split("]").str[0]
        fm_df["PredictedWinner"] = fm_df["Prediction"].str.extract(r"^(.+?) \d+-\d+")[0]
        fm_df["PredictedScore"] = fm_df["Prediction"].str.extract(r" (\d+-\d+)")[0]
        fm_df["WinProbability"] = fm_df["Prediction"].str.extract(r"\((\d+%)\)")[0]
        fm_df["PredictedPossessions"] = fm_df["Prediction"].str.extract(r"\[(\d+)\]")[0].astype(float)
        fm_df["Possessions"] = fm_df["Possessions"].where(fm_df["Possessions"] != "", fm_df["PredictedPossessions"])

        fm_df["PredictedMOV"] = [(int(x[0]) - int(x[1])) if len(x) > 1 else float("nan") for x 
                                 in fm_df.PredictedScore.astype("str").str.split("-")]

        fm_df.drop(["Prediction", "Time (ET)"], axis = 1, inplace = True)
        
        # Parse predicted loser.
        teams = fm_df.Game.str.split(", ").tolist()
        teams_np = fm_df.Game.str.split(" at ").tolist()
        pred_winner = fm_df["PredictedWinner"].tolist()
        
        i = 0
        pred_loser = []
        for i, x in enumerate(teams):
            if len(x) != 2:
                x = teams_np[i]
                
                # Account for neutral games.
                if len(x) < 2:
                    x = x[0].split(" vs. ")
                x[0] = " ".join(x[0].split()[1:])
                x[1] = " ".join(x[1].split()[1:])
                
            else:
                x[1] = x[1].split(" (")[0]

                x[0] = " ".join(x[0].split()[1:-1])
                x[1] = " ".join(x[1].split()[1:-1])
            
            if x[0] != pred_winner[i]:
                pred_loser.append(x[0])
            else:
                pred_loser.append(x[1])
                
            i = i + 1
            
        fm_df["PredictedLoser"] = pred_loser
        
        winner = fm_df.Game.str.split(", ").str[0].tolist()
        loser = fm_df.Game.str.split(", ").str[1].tolist()
        
        if not all(pd.isnull(loser)):
            loser = [str(x).split("(")[0] for x in loser]
            ot = fm_df.Game.str.split("(").str[1].astype("str").str.strip(")").tolist()
            fm_df["OT"] = ot
            
            fm_df["Loser"] = [" ".join(x.split()[1:-1]) if len(x.split(" at ")) < 2 else float("nan") for x in loser]
            fm_df["LoserRank"] = [x.split()[0] for x in loser]
            fm_df["LoserScore"] = [x.split()[-1] if len(x.split(" at ")) < 2 else float("nan") for x in loser]
            
        else:
            fm_df["OT"] = float("nan")
            fm_df["Loser"] = float("nan")
            fm_df["LoserRank"] = float("nan")
            fm_df["LoserScore"] = float("nan")
        
        fm_df["Winner"] = [" ".join(x.split()[1:-1]) if (len(x.split(" at ")) < 2 and 
            len(x.split(" vs. ")) < 2) else float("nan") for x in winner]
        fm_df["WinnerRank"] = [x.split()[0] if (len(x.split(" at ")) < 2 and 
            len(x.split(" vs. ")) < 2) else float("nan") for x in winner]
        fm_df["WinnerScore"] = [x.split()[-1] if (len(x.split(" at ")) < 2 and 
            len(x.split(" vs. ")) < 2) else float("nan") for x in winner]
        
        if not all(pd.isnull(loser)):
            fm_df["ActualMOV"] = [(int(x[0]) - int(x[1])) if not pd.isnull(x[0]) else float("nan") for x 
                                  in list(zip(fm_df.WinnerScore.tolist(), fm_df.LoserScore.tolist()))]
        else:
            fm_df["ActualMOV"] = float("nan")
        
        self.fm_df = fm_df



"""
This module contains functions for scraping the conference page kenpom.com tables into
pandas dataframes
"""

import pandas as pd
from io import StringIO
from cloudscraper import CloudScraper
from bs4 import BeautifulSoup
from typing import Optional
from .utils import get_html

def get_valid_conferences(browser: CloudScraper, season: Optional[str]=None):
	"""
	Scrapes the conferences (https://kenpom.com/conf.php) into a list.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.

	Returns:
		conference_list (list): List containing all valid conferences for the given season on kenpom.com.
	"""

	url = "https://kenpom.com/conf.php"
	url = url + '?c=B10'
	if(season):
		url = url + '&y=' + str(season)
	confs = BeautifulSoup(get_html(browser, url), "html.parser")
	table = confs.find_all('table')[-1]
	links = table.find_all('a')
	conf_list = []
	for link in links:
		conf_list.append(link['href'].split('=')[-1])
	conf_list.sort()
	return conf_list


def get_aggregate_stats(browser: CloudScraper, conf: Optional[str]=None, season: Optional[str]=None):
	"""
	Scrapes a given conference's stats (https://kenpom.com/conf.php or https://kenpom.com/confstats.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function
		conf (str: optional): conference abbreviation (ie B10, P12). If None, it will grab the table from https://kenpom.com/confstats.php instead of https://kenpom.com/conf.php
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.

	Returns:
		conference_df (dataframe): Dataframe containing aggregate stats of the conference for the given season on kenpom.com.
	"""
	if(conf):
		url = "https://kenpom.com/conf.php"
		url = url + f'?c={conf}'
		if(season):
			url = url + '&y=' + str(season)
		confs = BeautifulSoup(get_html(browser, url), "html.parser")
		#get first table
		table = confs.find_all('table')[-3]
		conf_df = pd.read_html(StringIO(str(table)))[0]
		#get second table
		table = confs.find_all('table')[-2]
		conf2_df = pd.read_html(StringIO(str(table)))[0]
		conf2_df['Value'] = conf2_df['Value'].str.replace('%', '').astype(float)
		conf_df = pd.concat([conf_df, conf2_df])
		#clean table
		conf_df = conf_df.set_index('Stat')
		conf_df = conf_df.drop('Unnamed: 1', axis=1)
		conf_df.columns = ['Value', 'Rank']
		conf_df.index = conf_df.index.str.split(' (', regex=False).str[0]
		return conf_df
	else:
		url = "https://kenpom.com/confstats.php"
		if(season):
			url = url + '?y=' + str(season)
		confs = BeautifulSoup(get_html(browser, url), "html.parser")
		#get table
		table = confs.find_all('table')[0]
		conf_df = pd.read_html(StringIO(str(table)))[0]
		# Clean table
		conf_df = conf_df.set_index('Conf')
		conf_df.columns = [stat[:-1] + 'Rank' if '.1' in stat else stat for stat in conf_df.columns]
		return conf_df

def get_standings(browser: CloudScraper, conf: str, season: Optional[str]=None):
	"""
	Scrapes a given conference's standing stats (https://kenpom.com/conf.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function
		conf (str): conference abbreviation (ie B10, P12)
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.

	Returns:
		conference_df (dataframe): Dataframe containing standing stats of the conference for the given season on kenpom.com.
	"""

	url = "https://kenpom.com/conf.php"
	url = url + f'?c={conf}'
	if(season):
		url = url + '&y=' + str(season)
	confs = BeautifulSoup(get_html(browser, url), "html.parser")
	table = confs.find_all('table')[0]
	conf_df = pd.read_html(StringIO(str(table)))[0]
	# Parse out seed
	conf_df['Seed'] = conf_df['Team'].str.extract('([0-9]+)')
	conf_df['Team'] = conf_df['Team'].str.replace('([0-9]+)', '', regex=True).str.rstrip()

	# Rename Rank headers
	conf_df.columns = [stat[:-1] + 'Rank' if '.1' in stat else stat for stat in conf_df.columns]

	return conf_df


def get_offense(browser: CloudScraper, conf: str, season: Optional[str]=None):
	"""
	Scrapes a given conference's offense only stats (https://kenpom.com/conf.php) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function
		conf (str): conference abbreviation (ie B10, P12)
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.

	Returns:
		conference_df (dataframe): Dataframe containing offensive stats of the conference for the given season on kenpom.com.
	"""

	url = "https://kenpom.com/conf.php"
	url = url + f'?c={conf}'
	if(season):
		url = url + '&y=' + str(season)
	confs = BeautifulSoup(get_html(browser, url), "html.parser")
	table = confs.find_all('table')[1]
	conf_df = pd.read_html(StringIO(str(table)))[0]

	# Rename Rank headers
	conf_df.columns = [stat[:-1] + 'Rank' if '.1' in stat else stat for stat in conf_df.columns]

	return conf_df


def get_defense(browser: CloudScraper, conf: str, season: Optional[str]=None):
	"""
	Scrapes a given conference's defense only stats (https://kenpom.com) into a dataframe.

	Args:
		browser (CloudScraper): Authenticated browser with full access to kenpom.com generated
			by the `login` function
		conf (str): conference abbreviation (ie B10, P12)
		season (str, optional): Used to define different seasons. 1999 is the earliest available season.

	Returns:
		conference_df (dataframe): Dataframe containing defensive stats of the conference for the given season on kenpom.com.
	"""

	url = "https://kenpom.com/conf.php"
	url = url + f'?c={conf}'
	if(season):
		url = url + '&y=' + str(season)
	confs = BeautifulSoup(get_html(browser, url), "html.parser")
	table = confs.find_all('table')[2]
	conf_df = pd.read_html(StringIO(str(table)))[0]

	# Rename Rank headers
	conf_df.columns = [stat[:-1] + 'Rank' if '.1' in stat else stat for stat in conf_df.columns]

	return conf_df
